/*
 * Copyright (c) 2002 Ian Dowse.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/sockio.h>
#include <sys/mbuf.h>
#include <sys/malloc.h>
#include <sys/kernel.h>
#include <sys/socket.h>

#include <net/if.h>
#include <net/if_arp.h>
#include <net/ethernet.h>
#include <net/if_dl.h>
#include <net/if_media.h>

#include <net/bpf.h>

#include <sys/bus.h>

#include <dev/usb/usb.h>
#include <dev/usb/usbdi.h>
#include <dev/usb/usbdi_util.h>
#include <dev/usb/usbdivar.h>
#include <dev/usb/usbdevs.h>
#include <dev/usb/usb_ethersubr.h>

#include "if_ruereg.h"

#define RUEFLG_GONE		0x01	/* Device no longer exists */
#define RUEFLG_IFINIT		0x02	/* Network interface registered */

#define ETHER_ALIGN		2

struct rue_type {
	u_int16_t vendor;
	u_int16_t product;
};

struct rue_softc {
	struct arpcom arpcom;
	struct ifmedia ifmedia;
	int unit;
	int flags;

	usbd_device_handle udev;
	usbd_interface_handle iface;
	int ed[RUE_ENDPT_MAX];
	usbd_pipe_handle ep[RUE_ENDPT_MAX];
	struct usb_qdat qdat;

	usbd_xfer_handle rx_xfer;
	struct mbuf *rx_m;

	usbd_xfer_handle tx_xfer;
	char tx_buf[1536];
	struct mbuf *tx_m;

	int if_flags;

	u_int16_t rxfilt;
};

Static struct rue_type rue_devs[] = {
	{ 0x0bda, 0x8150 },
	{ 0, 0 }
};

Static int rue_match(device_t);
Static int rue_attach(device_t);
Static int rue_detach(device_t);

Static void rue_clearstall(struct rue_softc *, usbd_xfer_handle, int);
Static void rue_clearstall_cb(usbd_xfer_handle, usbd_private_handle,
    usbd_status status);
Static struct mbuf *rue_newbuf(struct rue_softc *, struct mbuf *);
Static void rue_rxeof(usbd_xfer_handle, usbd_private_handle, usbd_status);
Static void rue_txeof(usbd_xfer_handle, usbd_private_handle, usbd_status);
Static void rue_rxstart(struct ifnet *);
Static void rue_start(struct ifnet *);
Static int rue_ioctl(struct ifnet *, u_long, caddr_t);
Static void rue_init(void *);
Static void rue_stop(struct rue_softc *);
Static void rue_watchdog(struct ifnet *);
Static void rue_shutdown(device_t);
Static int rue_ifmedia_upd(struct ifnet *);
Static void rue_ifmedia_sts(struct ifnet *, struct ifmediareq *);

Static void rue_setmulti(struct rue_softc *);
Static void rue_reset(struct rue_softc *);

Static usbd_status rue_blocking_readreg(struct rue_softc *, int, int, void *);
Static usbd_status rue_blocking_writereg(struct rue_softc *, int, int, void *);

Static device_method_t rue_methods[] = {
	/* Device interface */
	DEVMETHOD(device_probe,		rue_match),
	DEVMETHOD(device_attach,	rue_attach),
	DEVMETHOD(device_detach,	rue_detach),
	DEVMETHOD(device_shutdown,	rue_shutdown),

	{ 0, 0 }
};

Static driver_t rue_driver = {
	"rue",
	rue_methods,
	sizeof(struct rue_softc)
};

Static devclass_t rue_devclass;

MODULE_DEPEND(if_kue, usb, 1, 1, 1);
DRIVER_MODULE(if_rue, uhub, rue_driver, rue_devclass, usbd_driver_load, 0);

Static usbd_status
rue_blocking_readreg(struct rue_softc *sc, int start, int len, void *buf)
{
	usb_device_request_t req;

	req.bmRequestType = UT_READ_VENDOR_DEVICE;
	req.bRequest = RUE_CMD_READREG;
	USETW(req.wValue, start);
	USETW(req.wIndex, 0);
	USETW(req.wLength, len);
	return usbd_do_request(sc->udev, &req, buf);
}

Static usbd_status
rue_blocking_writereg(struct rue_softc *sc, int start, int len, void *buf)
{
	usb_device_request_t req;

	req.bmRequestType = UT_WRITE_VENDOR_DEVICE;
	req.bRequest = RUE_CMD_READREG;
	USETW(req.wValue, start);
	USETW(req.wIndex, 0);
	USETW(req.wLength, len);
	return usbd_do_request(sc->udev, &req, buf);
}

Static void
rue_reset(struct rue_softc *sc)
{
	usbd_status err;
	uByte cmdreg;
	int i;

	if (sc->flags & RUEFLG_GONE)
		return;

	/* Request a reset. */
	cmdreg = RUE_CR_SOFTRESET | RUE_CR_AUTOLOAD;
	err = rue_blocking_writereg(sc, RUE_CR, sizeof(cmdreg), &cmdreg);
	if (err)
		goto errout;

	/* Wait for the reset to complete. */
	for (i = 50; i > 0; i--) {
		err = rue_blocking_readreg(sc, RUE_CR, sizeof(cmdreg), &cmdreg);
		if (err)
			goto errout;
		if ((cmdreg & (RUE_CR_SOFTRESET | RUE_CR_AUTOLOAD)) == 0)
			break;
		tsleep(&i, PZERO, "rue_reset", (hz + 9) / 10);
	}
	if (i == 0)
		printf("rue%d: reset failed to complete\n", sc->unit);
        return;

errout:
	printf("rue%d: reset failed: %s\n", sc->unit, usbd_errstr(err));
}

USB_MATCH(rue)
{
	USB_MATCH_START(rue, uaa);
	struct rue_type *t;

	if (!uaa->iface)
		return(UMATCH_NONE);
	for (t = rue_devs; t->vendor != 0; t++)
		if (uaa->vendor == t->vendor && uaa->product == t->product)
			return (UMATCH_VENDOR_PRODUCT);

	return (UMATCH_NONE);
}

USB_ATTACH(rue)
{
	USB_ATTACH_START(rue, sc, uaa);
	char devinfo[1024];
	struct ifnet *ifp;
	usb_interface_descriptor_t *id;
	usb_endpoint_descriptor_t *ed;
	usbd_status err;
	int i, ep_idx;

	bzero(sc, sizeof(struct rue_softc));
	sc->iface = uaa->iface;
	sc->udev = uaa->device;
	sc->unit = device_get_unit(self);

	id = usbd_get_interface_descriptor(uaa->iface);

	usbd_devinfo(uaa->device, 0, devinfo);
	device_set_desc_copy(self, devinfo);
	printf("%s: %s\n", USBDEVNAME(self), devinfo);

	/* Reset the device. */
	rue_reset(sc);

	/* Find endpoints and open pipes. */
	for (i = 0; i < id->bNumEndpoints; i++) {
		ed = usbd_interface2endpoint_descriptor(uaa->iface, i);
		if (!ed) {
			printf("rue%d: couldn't get ep %d\n", sc->unit, i);
			USB_ATTACH_ERROR_RETURN;
		}
		ep_idx = -1;
		switch (ed->bmAttributes & UE_XFERTYPE) {
		case UE_BULK:
			if (UE_GET_DIR(ed->bEndpointAddress) == UE_DIR_IN)
				ep_idx = RUE_ENDPT_RX;
			else
				ep_idx = RUE_ENDPT_TX;
			break;
		case UE_INTERRUPT:
			if (UE_GET_DIR(ed->bEndpointAddress) == UE_DIR_IN)
				ep_idx = RUE_ENDPT_INTR;
			break;
		}
		if (ep_idx == -1)
			continue;
		if (sc->ed[ep_idx] != NULL) {
			printf("rue%d: duplicate matching endpoint\n",
			    sc->unit);
			continue;
		}

		/* Open the pipe. */
		sc->ed[ep_idx] = ed->bEndpointAddress;
		err = usbd_open_pipe(sc->iface, sc->ed[ep_idx],
		    USBD_EXCLUSIVE_USE, &sc->ep[ep_idx]);
		if (err) {
			printf("rue%d: open pipe %d failed: %s\n", sc->unit,
			    ed->bEndpointAddress, usbd_errstr(err));
			rue_detach(self);
			USB_ATTACH_ERROR_RETURN;
		}
	}
	/* Check that we got all pipes. */
	if (sc->ep[RUE_ENDPT_RX] == NULL || sc->ep[RUE_ENDPT_TX] == NULL ||
	    sc->ep[RUE_ENDPT_INTR] == NULL) {
		printf("rue%d: some pipes not found\n", sc->unit);
		rue_detach(self);
		USB_ATTACH_ERROR_RETURN;
	}

	/* Allocate xfer handles */
	sc->rx_xfer = usbd_alloc_xfer(uaa->device);
	sc->tx_xfer = usbd_alloc_xfer(uaa->device);
	if (sc->rx_xfer == NULL || sc->tx_xfer == NULL) {
		printf("rue%d: alloc xfer handles failed\n", sc->unit);
		rue_detach(self);
		USB_ATTACH_ERROR_RETURN;
	}

	/* Get ethernet address */
	if (rue_blocking_readreg(sc, RUE_IDR, ETHER_ADDR_LEN,
	    &sc->arpcom.ac_enaddr) != USBD_NORMAL_COMPLETION) {
		printf("rue%d: failed to read MAC address\n", sc->unit);
		USB_ATTACH_ERROR_RETURN;
	}
	printf("rue%d: Ethernet address: %6D\n", sc->unit,
	    (u_char *)&sc->arpcom.ac_enaddr, ":");

	ifp = &sc->arpcom.ac_if;
	ifp->if_softc = sc;
	ifp->if_unit = sc->unit;
	ifp->if_name = "rue";
	ifp->if_mtu = ETHERMTU;
	ifp->if_flags = IFF_BROADCAST | IFF_SIMPLEX | IFF_MULTICAST;
	ifp->if_ioctl = rue_ioctl;
	ifp->if_output = ether_output;
	ifp->if_start = rue_start;
	ifp->if_watchdog = rue_watchdog;
	ifp->if_init = rue_init;
	ifp->if_baudrate = 10000000;
	ifp->if_snd.ifq_maxlen = IFQ_MAXLEN;

	sc->qdat.ifp = ifp;
	sc->qdat.if_rxstart = rue_rxstart;

	/* Do ifmedia setup. */
	ifmedia_init(&sc->ifmedia, 0, rue_ifmedia_upd, rue_ifmedia_sts);
	ifmedia_add(&sc->ifmedia, IFM_ETHER | IFM_10_T, 0, NULL);
	ifmedia_add(&sc->ifmedia, IFM_ETHER | IFM_10_T | IFM_FDX, 0, NULL);
	ifmedia_add(&sc->ifmedia, IFM_ETHER | IFM_100_TX, 0, NULL);
	ifmedia_add(&sc->ifmedia, IFM_ETHER | IFM_100_TX | IFM_FDX, 0, NULL);
	ifmedia_add(&sc->ifmedia, IFM_ETHER | IFM_AUTO, 0, NULL);

	ifmedia_set(&sc->ifmedia, IFM_ETHER | IFM_AUTO);

	/* Call MI attach routine. */
	ether_ifattach(ifp, ETHER_BPF_SUPPORTED);
	sc->flags |= RUEFLG_IFINIT;
	usb_register_netisr();

	USB_ATTACH_SUCCESS_RETURN;
}

/*
 * Deallocate everything. This is designed to be safe to call from
 * rue_attach to clean up resources after a failure.
 */
USB_DETACH(rue)
{
	USB_DETACH_START(rue, sc);
	struct ifnet *ifp;
	int i;

	ifp = &sc->arpcom.ac_if;

	sc->flags |= RUEFLG_GONE;
	if (sc->flags & RUEFLG_IFINIT) {
		ether_ifdetach(ifp, ETHER_BPF_SUPPORTED);
		ifmedia_removeall(&sc->ifmedia);
	}

	for (i = 0; i < RUE_ENDPT_MAX; i++) {
		if (sc->ep[i] != NULL) {
			usbd_abort_pipe(sc->ep[i]);
			usbd_close_pipe(sc->ep[i]);
		}
	}

	if (sc->rx_xfer != NULL)
		usbd_free_xfer(sc->rx_xfer);
	if (sc->tx_xfer != NULL)
		usbd_free_xfer(sc->tx_xfer);

	return (0);
}

Static void
rue_setmulti(struct rue_softc *sc)
{
}

Static struct mbuf *
rue_newbuf(struct rue_softc *sc, struct mbuf *m)
{
	if (m == NULL) {
		MGETHDR(m, M_DONTWAIT, MT_DATA);
		if (m == NULL)
			return (NULL);
		MCLGET(m, M_DONTWAIT);
		if ((m->m_flags & M_EXT) == 0) {
			m_freem(m);
			return (NULL);
		}
	} else {
		m->m_data = m->m_ext.ext_buf;
	}

	m->m_len = m->m_pkthdr.len = MCLBYTES;
	m_adj(m, ETHER_ALIGN);
	return (m);
}

Static void
rue_rxstart(struct ifnet *ifp)
{
	struct rue_softc *sc;
	struct mbuf *m;

	sc = ifp->if_softc;

	KASSERT(sc->rx_m == NULL, ("rue_rxstart: already active"));
	if ((m = rue_newbuf(sc, NULL)) == NULL) {
		ifp->if_ierrors++;
		return;
	}

	/* Setup the transfer. */
	sc->rx_m = m;
	usbd_setup_xfer(sc->rx_xfer, sc->ep[RUE_ENDPT_RX], sc, mtod(m, char *),
	    m->m_len, USBD_SHORT_XFER_OK, USBD_NO_TIMEOUT, rue_rxeof);
	usbd_transfer(sc->rx_xfer);
}

/*
 * A frame has been uploaded: pass the resulting mbuf chain up to
 * the higher level protocols.
 */
Static void
rue_rxeof(usbd_xfer_handle xfer, usbd_private_handle priv, usbd_status status)
{
	struct rue_softc *sc;
        struct mbuf *m;
        struct ifnet *ifp;
	int len;
	uWord rxhdr;

	sc = priv;
	ifp = &sc->arpcom.ac_if;

	m = sc->rx_m;
	sc->rx_m = NULL;

	if (!(ifp->if_flags & IFF_RUNNING)) {
		m_freem(m);
		return;
	}

	if (status != USBD_NORMAL_COMPLETION) {
		m_freem(m);
		if (status == USBD_NOT_STARTED || status == USBD_CANCELLED)
			return;
		printf("rue%d: usb error on rx: %s\n", sc->unit,
		    usbd_errstr(status));
		if (status != USBD_STALLED)
			return;

		rue_clearstall(sc, xfer, sc->ed[RUE_ENDPT_RX]);
		return;
	}
	usbd_get_xfer_status(xfer, NULL, NULL, &len, NULL);

	if (len < ETHER_CRC_LEN) {
		ifp->if_ierrors++;
		goto done;
	}
	bcopy(mtod(m, char *) + len - ETHER_CRC_LEN, &rxhdr, sizeof(rxhdr));
	if ((UGETW(rxhdr) & RUE_RXHDR_VALID) == 0 ||
	    len < (UGETW(rxhdr) & RUE_RXHDR_BYTECNTMASK)) {
		ifp->if_ierrors++;
		printf("len %x, rxhdr %x\n", len, (int)UGETW(rxhdr));
		goto done;
	}

	len = (UGETW(rxhdr) & RUE_RXHDR_BYTECNTMASK) - ETHER_CRC_LEN;
	if (len < sizeof(struct ether_header)) {
		ifp->if_ierrors++;
		goto done;
	}

	ifp->if_ipackets++;
	m->m_pkthdr.rcvif = (struct ifnet *)&sc->qdat;
	m->m_pkthdr.len = m->m_len = len;

	/* Put the packet on the special USB input queue. */
	usb_ether_input(m);

	return;
done:
	/* Setup new transfer with the same mbuf. */
	sc->rx_m = m;
	usbd_setup_xfer(sc->rx_xfer, sc->ep[RUE_ENDPT_RX], sc, mtod(m, char *),
	    m->m_len, USBD_SHORT_XFER_OK, USBD_NO_TIMEOUT, rue_rxeof);
	usbd_transfer(sc->rx_xfer);
}

/*
 * The transmit frame download has completed.
 */
Static void
rue_txeof(usbd_xfer_handle xfer, usbd_private_handle priv, usbd_status status)
{
	struct rue_softc *sc;
	struct ifnet *ifp;
	struct mbuf *m;
	int s;

	sc = priv;
	ifp = &sc->arpcom.ac_if;

	m = sc->tx_m;
	sc->tx_m = NULL;

	s = splimp();
	ifp->if_timer = 0;
	ifp->if_flags &= ~IFF_OACTIVE;
	splx(s);
	if (status != USBD_NORMAL_COMPLETION) {
		m_freem(m);
		ifp->if_oerrors++;
		if (status == USBD_NOT_STARTED || status == USBD_CANCELLED)
			return;
		printf("rue%d: usb error on tx: %s\n", sc->unit,
		    usbd_errstr(status));
		if (status == USBD_STALLED)
			rue_clearstall(sc, xfer, sc->ed[RUE_ENDPT_TX]);
		return;
	} else {
		ifp->if_opackets++;
	}

	/* The used mbuf becomes a token to initiate the next transmit. */
	m->m_pkthdr.rcvif = ifp;
	usb_tx_done(m);
}

Static void
rue_clearstall(struct rue_softc *sc, usbd_xfer_handle xfer, int endpt_descr)
{
	usb_device_request_t req;
	usbd_status err;

	bzero(&req, sizeof(req));
	req.bmRequestType = UT_WRITE_ENDPOINT;
	req.bRequest = UR_CLEAR_FEATURE;
	USETW(req.wValue, UF_ENDPOINT_HALT);
	USETW(req.wIndex, endpt_descr);
	USETW(req.wLength, 0);
	usbd_setup_default_xfer(xfer, sc->udev, sc, 5000, &req, NULL, 0, 0,
	    rue_clearstall_cb);
	err = usbd_transfer(xfer);
	if (err && err != USBD_IN_PROGRESS)
		printf("rue%d: clearstall setup failed: %s\n", sc->unit,
		    usbd_errstr(err));
}

Static void
rue_clearstall_cb(usbd_xfer_handle xfer, usbd_private_handle priv,
    usbd_status status) {
	struct rue_softc *sc = priv;

	if (status != USBD_NORMAL_COMPLETION) {
		printf("rue%d: clearstall failed: %s\n", sc->unit,
		    usbd_errstr(status));
		return;
	}
	rue_start(&sc->arpcom.ac_if);
	if (sc->rx_m == NULL)
		rue_rxstart(&sc->arpcom.ac_if);
}

Static void
rue_start(struct ifnet *ifp)
{
	struct rue_softc *sc;
	struct mbuf *m;
	usbd_status err;
	int len;

	sc = ifp->if_softc;

	if (ifp->if_flags & IFF_OACTIVE)
		return;

	IF_DEQUEUE(&ifp->if_snd, m);
	if (m == NULL)
		return;

	len = m->m_pkthdr.len;
	if (len > sizeof(sc->tx_buf))
		len = sizeof(sc->tx_buf);
	m_copydata(m, 0, len, sc->tx_buf);

	usbd_setup_xfer(sc->tx_xfer, sc->ep[RUE_ENDPT_TX], sc, sc->tx_buf, len,
	    0, 10000, rue_txeof);
	err = usbd_transfer(sc->tx_xfer);
	if (err != USBD_IN_PROGRESS) {
		printf("rue%d: usb error in rue_start: %s\n", sc->unit,
		    usbd_errstr(err));
		m_freem(m);
		return;
	}
	sc->tx_m = m;

	if (ifp->if_bpf)
		bpf_mtap(ifp, m);

	ifp->if_flags |= IFF_OACTIVE;
	ifp->if_timer = 5;
}

Static void
rue_init(void *xsc)
{
	struct rue_softc *sc = xsc;
	struct ifnet *ifp = &sc->arpcom.ac_if;
	uByte reg8;
	uWord reg16;
	int s;

	if (ifp->if_flags & IFF_RUNNING)
		return;

	s = splimp();

	/* Set MAC address */
	rue_blocking_writereg(sc, RUE_IDR, ETHER_ADDR_LEN,
	    &sc->arpcom.ac_enaddr);

	/* Enable receiver and transmitter. */
	rue_blocking_readreg(sc, RUE_CR, sizeof(reg8), &reg8);
	reg8 |= RUE_CR_TE | RUE_CR_RE;
	rue_blocking_writereg(sc, RUE_CR, sizeof(reg8), &reg8);

	/* Set up the receiver. */
	rue_blocking_readreg(sc, RUE_RCR, sizeof(reg16), &reg16);
	USETW(reg16, UGETW(reg16) | RUE_RCR_TAIL | RUE_RCR_AD | RUE_RCR_AB |
	    RUE_RCR_AM);
	if (ifp->if_flags & IFF_PROMISC)
		USETW(reg16, UGETW(reg16) | RUE_RCR_AAP);
	rue_blocking_writereg(sc, RUE_RCR, sizeof(reg16), &reg16);

	/* Load the multicast filter. */
	rue_setmulti(sc);

	/* Start receiving if necessary */
	if (sc->rx_m == NULL)
		rue_rxstart(ifp);

	ifp->if_flags |= IFF_RUNNING;
	ifp->if_flags &= ~IFF_OACTIVE;

	splx(s);
}

Static int
rue_ifmedia_upd(struct ifnet *ifp)
{
	struct rue_softc *sc = ifp->if_softc;
	int media;
	uWord reg16;

	rue_blocking_readreg(sc, RUE_BMCR, sizeof(reg16), &reg16);
	USETW(reg16, UGETW(reg16) & ~(RUE_BMCR_FULLDUPLEX | RUE_BMCR_INITNEG |
	    RUE_BMCR_ANE | RUE_BMCR_SPEED100 | RUE_BMCR_RESET));

	media = sc->ifmedia.ifm_media;
	switch(IFM_SUBTYPE(media)) {
	case IFM_100_TX:
		USETW(reg16, UGETW(reg16) | RUE_BMCR_SPEED100);
		/* fallthrough */
	case IFM_10_T:
		if (media & IFM_FDX)
			USETW(reg16, UGETW(reg16) | RUE_BMCR_FULLDUPLEX);
		break;
	default:
		USETW(reg16, UGETW(reg16) | RUE_BMCR_INITNEG | RUE_BMCR_ANE);
		break;
	}

	rue_blocking_writereg(sc, RUE_BMCR, sizeof(reg16), &reg16);
	return (0);
}

Static void
rue_ifmedia_sts(struct ifnet *ifp, struct ifmediareq *ifmr)
{
	struct rue_softc *sc = ifp->if_softc;
	uWord reg16;

	rue_blocking_readreg(sc, RUE_MSR, sizeof(reg16), &reg16);
	ifmr->ifm_status = IFM_AVALID;
	ifmr->ifm_active = IFM_ETHER;
	if (UGETW(reg16) & RUE_MSR_LINK)
		ifmr->ifm_status |= IFM_ACTIVE;
	if (UGETW(reg16) & RUE_MSR_SPEED100)
		ifmr->ifm_active |= IFM_100_TX;
	else
		ifmr->ifm_active |= IFM_10_T;
	if (UGETW(reg16) & RUE_MSR_DUPLEX)
		ifmr->ifm_active |= IFM_FDX;
}

Static int
rue_ioctl(struct ifnet *ifp, u_long command, caddr_t data)
{
	struct rue_softc *sc = ifp->if_softc;
	struct ifreq *ifr = (struct ifreq *)data;
	int s, error = 0;

	s = splimp();

	switch(command) {
	case SIOCSIFADDR:
	case SIOCGIFADDR:
	case SIOCSIFMTU:
		error = ether_ioctl(ifp, command, data);
		break;
	case SIOCSIFFLAGS:
		if (ifp->if_flags & IFF_UP) {
			if (!(ifp->if_flags & IFF_RUNNING))
				rue_init(sc);
		} else {
			if (ifp->if_flags & IFF_RUNNING)
				rue_stop(sc);
		}
		sc->if_flags = ifp->if_flags;
		error = 0;
		break;
	case SIOCADDMULTI:
	case SIOCDELMULTI:
		rue_setmulti(sc);
		error = 0;
		break;
	case SIOCGIFMEDIA:
	case SIOCSIFMEDIA:
		error = ifmedia_ioctl(ifp, ifr, &sc->ifmedia, command);
		break;
	default:
		error = EINVAL;
		break;
	}

	(void)splx(s);

	return(error);
}

Static void
rue_watchdog(struct ifnet *ifp)
{
	struct rue_softc *sc;

	sc = ifp->if_softc;

	ifp->if_oerrors++;
	printf("rue%d: watchdog timeout\n", sc->unit);
}

Static void
rue_stop(struct rue_softc *sc)
{
	usbd_status err;
	struct ifnet *ifp;
	int i, s;

	s = splimp();
	ifp = &sc->arpcom.ac_if;
	ifp->if_timer = 0;

	/* Stop transfers. */
	for (i = 0; i < RUE_ENDPT_MAX; i++) {
		err = usbd_abort_pipe(sc->ep[i]);
		if (err)
			printf("rue%d: abort pipe %d failed: %s\n", sc->unit, i,
			    usbd_errstr(err));
	}

	/* Free any mbufs */
	if (sc->rx_m != NULL) {
		m_freem(sc->rx_m);
		sc->rx_m = NULL;
	}
	if (sc->tx_m != NULL) {
		m_freem(sc->tx_m);
		sc->tx_m = NULL;
	}
	rue_reset(sc);

	ifp->if_flags &= ~(IFF_RUNNING | IFF_OACTIVE);
	splx(s);
}

Static void
rue_shutdown(device_t dev)
{
	struct rue_softc *sc;

	sc = device_get_softc(dev);
	rue_stop(sc);
}
